"""
Simple tests for python harness

Usage:
$ pytest test_harness.py

"""
from crawleruseragents import is_crawler, matching_crawlers


def test_match():
    assert is_crawler("test Googlebot/2.0 test") is True


def test_nomatch():
    assert is_crawler("!!!!!!!!!!!!") is False


def test_case():
    assert is_crawler("test googlebot/2.0 test") is False


def test_matching_crawlers_match():
    result = matching_crawlers("test Googlebot/2.0 test")
    assert isinstance(result, list)
    assert len(result) > 0
    assert all(isinstance(val, int) for val in result)


def test_matching_crawlers_nomatch():
    result = matching_crawlers("!!!!!!!!!!!!")
    assert isinstance(result, list)
    assert len(result) == 0


def test_matching_crawlers_case():
    result = matching_crawlers("test googlebot/2.0 test")
    assert isinstance(result, list)
    assert len(result) == 0



import re
import json
from pathlib import Path


def load_json():
    cwd = Path(__file__).parent
    user_agents_file_path = cwd / "crawler-user-agents.json"
    with user_agents_file_path.open() as patterns_file:
        return json.load(patterns_file)


CRAWLER_USER_AGENTS_DATA = load_json()
CRAWLER_USER_AGENTS_REGEXP = re.compile(
    "|".join(i["pattern"] for i in CRAWLER_USER_AGENTS_DATA)
)


def is_crawler(user_agent: str) -> bool:
    """Return True if the given User-Agent matches a known crawler."""
    return bool(CRAWLER_USER_AGENTS_REGEXP.search(user_agent))


def matching_crawlers(user_agent: str) -> list[int]:
    """
    Return a list of the indices in CRAWLER_USER_AGENTS_DATA of any crawlers
    matching the given User-Agent.
    """
    result = []
    if is_crawler(user_agent):
        for num, crawler_user_agent in enumerate(CRAWLER_USER_AGENTS_DATA):
            if re.search(crawler_user_agent["pattern"], user_agent):
                result.append(num)
    return result



"""
Break crawler-user-agents.json in ways that validate.py should detect

Usage:
$ pytest test_validation.py

"""
from __future__ import print_function

import json
import subprocess
from copy import deepcopy

import pytest


def update_json_file(data):
    with open('crawler-user-agents.json', 'w') as f:
        json.dump(data, f, indent=2)

@pytest.fixture
def restore_original_json():
    # Load original version of crawler-user-agents.json
    with open('crawler-user-agents.json') as f:
        original_json = json.load(f)

    # By using a yield statement instead of return, all the code after
    # the yield statement serves as the teardown code:
    yield None

    # tear down code: restore original version of crawler-user-agents.json
    update_json_file(original_json)


def assert_validate_failed():
    assert subprocess.call(['python', 'validate.py']) != 0


def assert_validate_passed():
    assert subprocess.call(['python', 'validate.py']) == 0


def test_simple_pass(restore_original_json):
    # the json must be an array of objects containing "pattern"
    # there must be more than 10 instances to pass
    user_agent_list = [{'pattern': 'foo',
                        'instances': ['foo',
                                      'afoo',
                                      'foob',
                                      'cfood',
                                      '/foo/',
                                      '\\foo\\',
                                      ':foo:',
                                      'foo.',
                                      '!foo',
                                      '/foo',
                                      'foo\\',
                                      'FoofooFoo',
                                      'foot']}]
    update_json_file(user_agent_list)
    assert_validate_passed()

def test_simplest_pass(restore_original_json):
    # the simplest crawler file passes
    user_agent_list = [{'pattern': 'foo', 'instances': []}]
    update_json_file(user_agent_list)
    assert_validate_passed()

def test_schema_violation_dict1(restore_original_json):
    # contract: the json must be an array
    user_agent_list = {'foo':None}
    update_json_file(user_agent_list)
    assert_validate_failed()

def test_schema_violation_dict2(restore_original_json):
    # contract: the json must be an array of objects containing "pattern"
    user_agent_list = [{'foo':None}]
    update_json_file(user_agent_list)
    assert_validate_failed()

def test_schema_violation_dict3(restore_original_json):
    # contract: the json must be an array of objects containing "pattern" and valid properties
    user_agent_list = [{'pattern':'foo', 'foo':3}]
    update_json_file(user_agent_list)
    assert_validate_failed()

def test_simple_duplicate_detection(restore_original_json):
    # contract: if we have the same pattern twice, it fails
    user_agent_list = [{'pattern': 'foo',
                        'instances': ['foo',
                                      'afoo',
                                      'foob',
                                      'cfood',
                                      '/foo/',
                                      '\\foo\\',
                                      ':foO:',
                                      'foo.',
                                      '!foo',
                                      '/foo',
                                      'foo\\',
                                      'FoofooFoo',
                                      'foot']},
                        {'pattern': 'foo'}]
    update_json_file(user_agent_list)
    assert_validate_failed()


def test_simple_duplicate_detection2(restore_original_json):
    # contract: if we have the same pattern twice, it fails (even w/o instances)
    user_agent_list = [{'pattern': 'foo',
                        'instances': ['foo',
                                      'afoo',
                                      'foob',
                                      'cfood',
                                      '/foo/',
                                      '\\foo\\',
                                      ':foo:',
                                      'foo.',
                                      '!foo',
                                      '/foo',
                                      'foo\\',
                                      'FoofooFoo',
                                      'foot']},
                       {'pattern': 'bar'},
                       {'pattern': 'bar'}]
    update_json_file(user_agent_list)
    assert_validate_failed()


def test_subset_duplicate_detection(restore_original_json):
    # contract: if a pattern matches another pattern, it fails
    user_agent_list = [{'pattern': 'foo',
                        'instances': ['foo',
                                      'afoo',
                                      'foob',
                                      'cfood',
                                      '/foo/',
                                      '\\foo\\',
                                      ':foo:',
                                      'foo.',
                                      '!foo',
                                      '/foo',
                                      'foo\\',
                                      'FoofooFoo',
                                      'foot']},
                       {'pattern': 'afoot'}]
    update_json_file(user_agent_list)
    assert_validate_failed()


def test_case_sensitivity(restore_original_json):
    # contract: the patterns are case sensitive
    user_agent_list = [{'pattern': 'foo',
                        'instances': ['FOO',
                                      'aFoo',
                                      'foob',
                                      'cfood',
                                      '/foo/',
                                      '\\foo\\',
                                      ':foo:',
                                      'fOo.',
                                      '!FOO',
                                      '/foo',
                                      'foo\\',
                                      'FoofooFoo',
                                      'foot']}]
    update_json_file(user_agent_list)
    assert_validate_failed()

def test_duplicate_case_insensitive_detection(restore_original_json):
    # contract: fail if we have patterns that differ only in capitailization
    user_agent_list = [{'pattern': 'foo',
                        'instances': ['foo',
                                      'afoo',
                                      'foob',
                                      'cfood',
                                      '/foo/',
                                      '\\foo\\',
                                      ':foo:',
                                      'foo.',
                                      '!foo',
                                      '/foo',
                                      'foo\\',
                                      'FoofooFoo',
                                      'foot']},
                       {'pattern': 'fOo'}]
    update_json_file(user_agent_list)
    assert_validate_failed()



"""
Validate JSON to ensure that patterns all work
"""
from __future__ import print_function

import json
import re
from collections import Counter
import datetime

from jsonschema import validate


JSON_SCHEMA = {
    "type": "array",
    "items": {
        "type": "object",
        "properties": {
            "pattern": {"type": "string"}, # required
            "instances": {"type": "array"}, # required
            "url": {"type": "string"}, # optional
            "description": {"type": "string"}, # optional
            "addition_date": {"type": "string"}, # optional
            "depends_on": {"type": "array"} # allows an instance to match twice
        },
        "required": ["pattern", "instances"]
    }
}


def main():
    with open('crawler-user-agents.json') as f:
        json_data = json.load(f)

    # check format using JSON Schema
    validate(json_data, JSON_SCHEMA)

    # check for simple duplicates
    pattern_counts = Counter(entry['pattern'] for entry in json_data)
    for pattern, count in pattern_counts.most_common():
        if count > 1:
            raise ValueError('Pattern {!r} appears {} times'.format(pattern,
                                                                    count))

    # check for duplicates with different capitalization
    pattern_counts = Counter(entry['pattern'].lower() for entry in json_data)
    for pattern, count in pattern_counts.most_common():
        if count > 1:
            raise ValueError('Pattern {!r} is duplicated {} times with different capitalization'
                             .format(pattern, count))
    
    # checks that no pattern contains unescaped slash / 
    for entry in json_data:
        pattern = entry['pattern']
        if re.search('[^\\\\]/', pattern):
            raise ValueError('Pattern {!r} has an unescaped slash character'.format(pattern))

    # check that no pattern contains unescaped dot .
    for entry in json_data:
        pattern = entry['pattern']
        if re.search('[^\\\\]\\.', pattern):
            raise ValueError('Pattern {!r} has an unescaped dot character'.format(pattern))

    # check that we match the given instances
    num_instances = 0
    for entry in json_data:
        pattern = entry['pattern']
        
        # assert that field "addition_date" has format "2019/12/23",
        if 'addition_date' in entry:
            if not re.match(r'\d{4}/\d{2}/\d{2}', entry['addition_date']):
                raise ValueError('addition_date {!r} has invalid format'.format(entry['addition_date']))
            # parse the date with datetime
            datetime.datetime.strptime(entry['addition_date'], '%Y/%m/%d')
            
        # canonicalize entry
        if 'depends_on' not in entry: entry['depends_on'] = []
            
        # check that we have only the rights properties (not handled by default in module jsonschema)
        assert set([str(x) for x in entry.keys()]).issubset(set(JSON_SCHEMA['items']['properties'].keys())), "the entry contains unknown properties"  
        instances = entry.get('instances')
        if instances:
            # check that there is no duplicate
            if not len(instances) == len(set(instances)):
                raise Exception("duplicate instances in "+pattern)
            for instance in instances:
                num_instances += 1
                if not re.search(pattern, instance):
                    raise ValueError('Pattern {!r} misses instance {!r}'
                                     .format(pattern, instance))
                

    # Make sure we have at least one pattern
    if len(json_data) < 1:
        raise Exception("no pattern")

    # Check for patterns that match other patterns
    for entry1 in json_data:
        for entry2 in json_data:
            if entry1 != entry2 and re.search(entry1['pattern'],
                                              entry2['pattern'],re.IGNORECASE):
                raise ValueError('Pattern {!r} is a subset of {!r}'
                                 .format(entry2['pattern'], entry1['pattern']))

    print('Validation passed')


if __name__ == '__main__':
    main()
